# -*- coding: utf-8 -*-
#
#    EAV-Django is a reusable Django application which implements EAV data model
#    Copyright © 2009—2010  Andrey Mikhaylenko
#
#    This file is part of EAV-Django.
#
#    EAV-Django is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    EAV-Django is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with EAV-Django.  If not, see <http://gnu.org/licenses/>.

from django import forms
from django.utils.translation import ugettext_lazy as _


class RangeWidget(forms.MultiWidget):
    def __init__(self, attrs=None):
        widgets = (forms.TextInput(attrs=attrs), forms.TextInput(attrs=attrs))
        attrs = attrs or {}
        attrs.update({'class': 'range'})
        super(RangeWidget, self).__init__(widgets, attrs)

    def decompress(self, value):
        return value or [None, None]

    def format_output(self, rendered_widgets):
        names = 'min max'.split()
        pairs = zip(names, rendered_widgets)
        return _('%(min)s to %(max)s') % dict(pairs)



# -*- coding: utf-8 -*-
#
#    EAV-Django is a reusable Django application which implements EAV data model
#    Copyright © 2009—2010  Andrey Mikhaylenko
#
#    This file is part of EAV-Django.
#
#    EAV-Django is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    EAV-Django is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with EAV-Django.  If not, see <http://gnu.org/licenses/>.

# django
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes import generic
from django.db.models import (BooleanField, CharField, DateField, FloatField,
                              ForeignKey, IntegerField, Model, NullBooleanField,
                              TextField)
from django.utils.translation import ugettext_lazy as _

# 3rd-party
from autoslug.fields import AutoSlugField
from autoslug.settings import slugify
#from view_shortcuts.decorators import cached_property

# this app
from managers import BaseEntityManager


__all__ = ['BaseAttribute', 'BaseChoice', 'BaseEntity', 'BaseSchema']


def slugify_attr_name(name):
    return slugify(name.replace('_', '-')).replace('-', '_')


def get_entity_lookups(entity):
    ctype = ContentType.objects.get_for_model(entity)
    return {'entity_type': ctype, 'entity_id': entity.pk}


class BaseSchema(Model):
    """
    Metadata for an attribute.
    """
    TYPE_TEXT    = 'text'
    TYPE_FLOAT   = 'float'
    TYPE_DATE    = 'date'
    TYPE_BOOLEAN = 'bool'
    TYPE_MANY    = 'many'
    TYPE_RANGE   = 'range'

    DATATYPE_CHOICES = (
        (TYPE_TEXT,    _('text')),
        (TYPE_FLOAT,   _('number')),
        (TYPE_DATE,    _('date')),
        (TYPE_BOOLEAN, _('boolean')),
        (TYPE_MANY,    _('multiple choices')),
        (TYPE_RANGE,   _('numeric range')),
    )

    title    = CharField(_('title'), max_length=100, help_text=_('user-friendly attribute name'))
    name     = AutoSlugField(_('name'), populate_from='title',
                             editable=True, blank=True, slugify=slugify_attr_name)
    help_text = CharField(_('help text'), max_length=250, blank=True,
                          help_text=_('short description for administrator'))
    datatype = CharField(_('data type'), max_length=5, choices=DATATYPE_CHOICES)

    required = BooleanField(_('required'))
    searched = BooleanField(_('include in search'))  # i.e. full-text search? mb for text only
    filtered = BooleanField(_('include in filters'))
    sortable = BooleanField(_('allow sorting'))

    class Meta:
        abstract = True
        verbose_name, verbose_name_plural = _('schema'), _('schemata')
        ordering = ['title']

    def __unicode__(self):
        return u'%s (%s)%s' % (self.title, self.get_datatype_display(),
                                u' %s'%_('required') if self.required else '')

    def get_choices(self, entity=None):
        """
        Returns a list of name/title tuples::

            schema.get_choices()    # --> [("green", "Green color"), ("red", "Red color")]

        Names are used for lookups, titles are displayed to user.

        This method must be overloaded by subclasses of BaseSchema to enable
        many-to-one schemata machinery.
        """
        return self.choices.all()

    def get_attrs(self, entity):
        """
        Returns available attributes for given entity instance.
        Handles many-to-one relations transparently.
        """
        return self.attrs.filter(**get_entity_lookups(entity))

    def save_attr(self, entity, value):
        """
        Saves given EAV attribute with given value for given entity.

        If schema is not many-to-one, the value is saved to the corresponding
        Attr instance (which is created or updated).

        If schema is many-to-one, the value is processed thusly:

        * if value is iterable, all Attr instances for corresponding managed m2m
          schemata are updated (those with names from the value list are set to
          True, others to False). If a list item is not in available choices,
          ValueError is raised;
        * if the value is None, all corresponding Attr instances are reset to False;
        * if the value is neither a list nor None, it is wrapped into a list and
          processed as above (i.e. "foo" --> ["foo"]).
        """

        if self.datatype == self.TYPE_MANY:
            self._save_m2m_attr(entity, value)
        else:
            self._save_single_attr(entity, value)

    def _save_single_attr(self, entity, value=None, schema=None,
                          create_nulls=False, extra={}):
        """
        Creates or updates an EAV attribute for given entity with given value.

        :param schema: schema for attribute. Default it current schema instance.
        :param create_nulls: boolean: if True, even attributes with value=None
            are created (by default they are skipped).
        :param extra: dict: additional data for Attr instance (e.g. title).
        """
        # If schema is not many-to-one, the value is saved to the corresponding
        # Attr instance (which is created or updated).

        schema = schema or self
        lookups = dict(get_entity_lookups(entity), schema=schema, **extra)
        try:
            attr = self.attrs.get(**lookups)
        except self.attrs.model.DoesNotExist:
            attr = self.attrs.model(**lookups)
        if create_nulls or value != attr.value:
            attr.value = value
            for k,v in extra.items():
                setattr(attr, k, v)
            attr.save()

    def _save_m2m_attr(self, entity, value):

        if not hasattr(value, '__iter__'):
            value = [value]

        if not all(isinstance(x, BaseChoice) for x in value):
            raise TypeError('Cannot assign "%s": "Attr.choice" '
                            'must be a BaseChoice instance.'
                            % ', '.join(value))

        # drop all attributes for this entity/schema pair
        self.get_attrs(entity).delete()

        # Attr instances for corresponding managed m2m schemata are updated
        for choice in value:
            self._save_single_attr(
                entity,
                schema = self,
                create_nulls = True,
                extra = {'choice': choice}
            )


class BaseEntity(Model):
    """
    Entity, the "E" in EAV. This model is abstract and must be subclassed.
    See tests for examples.
    """

    objects = BaseEntityManager()

    class Meta:
        abstract = True

    def save(self, force_eav=False, **kwargs):
        """
        Saves entity instance and creates/updates related attribute instances.

        :param eav: if True (default), EAV attributes are saved along with entity.
        """
        # save entity
        super(BaseEntity, self).save(**kwargs)

        # TODO: think about use cases; are we doing it right?
        #if not self.check_eav_allowed():
        #    import warnings
        #    warnings.warn('EAV attributes are going to be saved along with entity'
        #                  ' despite %s.check_eav_allowed() returned False.'
        #                  % type(self), RuntimeWarning)


        # create/update EAV attributes
        for schema in self.get_schemata():
            value = getattr(self, schema.name, None)
            schema.save_attr(self, value)

    def __getattr__(self, name):
        if not name.startswith('_'):
            if name in self.get_schema_names():
                schema = self.get_schema(name)
                attrs = schema.get_attrs(self)
                if schema.datatype == schema.TYPE_MANY:
                    return [a.value for a in attrs if a.value]
                else:
                    return attrs[0].value if attrs else None
        raise AttributeError('%s does not have attribute named "%s".' %
                             (self._meta.object_name, name))

    def __iter__(self):
        "Iterates over non-empty EAV attributes. Normal fields are not included."
        for attr in self.attrs.select_related():
            if getattr(self, attr.schema.name, None):
                yield attr

    @classmethod
    def get_schemata_for_model(cls):
        return NotImplementedError('BaseEntity subclasses must define method '
                                   '"get_schemata_for_model" which returns a '
                                   'QuerySet for a BaseSchema subclass.')

    def get_schemata_for_instance(self, qs):
        return qs

    def get_schemata(self):
        if hasattr(self, '_schemata_cache') and self._schemata_cache is not None:
            return self._schemata_cache
        all_schemata = self.get_schemata_for_model().select_related()
        self._schemata_cache = self.get_schemata_for_instance(all_schemata)
        self._schemata_cache_dict = dict((s.name, s) for s in self._schemata_cache)
        return self._schemata_cache

    def get_schema_names(self):
        if not hasattr(self, '_schemata_cache_dict'):
            self.get_schemata()
        return self._schemata_cache_dict.keys()

    def get_schema(self, name):
        if not hasattr(self, '_schemata_cache_dict'):
            self.get_schemata()
        return self._schemata_cache_dict[name]

    def get_schema_by_id(self, schema_id):
        for schema in self.get_schemata():
            if schema.pk == schema_id:
                return schema

    def check_eav_allowed(self):
        """
        Returns True if entity instance allows EAV attributes to be attached.

        Can be useful if some external data is required to determine available
        schemata and that data may be missing. In such cases this method should
        be overloaded to check whether the data is available.
        """
        return True

    def is_valid(self):
        "Returns True if attributes and their values conform with schema."

        raise NotImplementedError()

        '''
        schemata = self.rubric.schemata.all()
        return all(x.is_valid for x in self.attributes)
        # 1. check if all required attributes are present
        for schema in schemata:
            pass
        # 2. check if all attributes have appropriate values
        for schema in schemata:
            pass
        return True
        '''


class BaseChoice(Model):
    title = CharField(max_length=100)
    schema = NotImplemented

    class Meta:
        abstract = True
        ordering = ('title',)

    def __unicode__(self):
        return self.title   #u'%s "%s"' % (self.schema.title, self.title)


class BaseAttribute(Model):
    entity_type = ForeignKey(ContentType)
    entity_id = IntegerField()
    entity = generic.GenericForeignKey(ct_field="entity_type", fk_field='entity_id')

    value_text = TextField(blank=True, null=True)
    value_float = FloatField(blank=True, null=True)
    value_date = DateField(blank=True, null=True)
    value_bool = NullBooleanField(blank=True)    # TODO: ensure that form invalidates null booleans (??)
    value_range_min = FloatField(blank=True, null=True)
    value_range_max = FloatField(blank=True, null=True)

    schema = NotImplemented    # must be FK
    choice = NotImplemented    # must be nullable FK

    class Meta:
        abstract = True
        verbose_name, verbose_name_plural = _('attribute'), _('attributes')
        ordering = ['entity_type', 'entity_id', 'schema']
        unique_together = ('entity_type', 'entity_id', 'schema', 'choice')

    def __unicode__(self):
        return u'%s: %s "%s"' % (self.entity, self.schema.title, self.value)

    def _get_value(self):
        if self.schema.datatype == self.schema.TYPE_MANY:
            return self.choice
        if self.schema.datatype == self.schema.TYPE_RANGE:
            names = ('value_range_%s' % x for x in ('min', 'max'))
            value = tuple(getattr(self, x, None) for x in names)
            return None if value == (None, None) else value
        return getattr(self, 'value_%s' % self.schema.datatype)

    def _set_value(self, new_value):
        if self.schema.datatype == self.schema.TYPE_RANGE:
            new_value = new_value or (None, None)

            # validate range value -- expecting a tuple of two numbers
            try:
                validate_range_value(new_value)
            except (TypeError, ValueError):
                raise

            for k,v in zip('min max'.split(), new_value):
                v = v if v is None else float(v)
                setattr(self, 'value_range_%s' % k, v)
        else:
            setattr(self, 'value_%s' % self.schema.datatype, new_value)

    value = property(_get_value, _set_value)


def validate_range_value(value):
    """
    Validates given value against `Schema.TYPE_RANGE` data type. Raises
    TypeError or ValueError if something is wrong. Returns None if everything
    is OK.
    """
    if value == (None, None):
        return

    if not hasattr(value, '__iter__'):
        raise TypeError('Range value must be an iterable, got "%s".' % value)
    if not 2 == len(value):
        raise ValueError('Range value must consist of two elements, got %d.' %
                         len(value))
    if not all(isinstance(x, (int,float)) for x in value):
        raise TypeError('Range value must consist of two numbers, got "%s" '
                        'and "%s" instead.' % value)
    if not value[0] <= value[1]:
        raise ValueError('Range must consist of min and max values (min <= '
                         'max) but got "%s" and "%s" instead.' % value)
    return


# xxx catch signal Attr.post_save() --> update attr.item.attribute_cache (JSONField or such)



# -*- coding: utf-8 -*-
#
#    EAV-Django is a reusable Django application which implements EAV data model
#    Copyright © 2009—2010  Andrey Mikhaylenko
#
#    This file is part of EAV-Django.
#
#    EAV-Django is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    EAV-Django is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with EAV-Django.  If not, see <http://gnu.org/licenses/>.

# django
from django.contrib.admin import helpers, ModelAdmin
from django.utils.safestring import mark_safe


class BaseEntityAdmin(ModelAdmin):

    def render_change_form(self, request, context, **kwargs):
        """
        Wrapper for ModelAdmin.render_change_form. Replaces standard static
        AdminForm with an EAV-friendly one. The point is that our form generates
        fields dynamically and fieldsets must be inferred from a prepared and
        validated form instance, not just the form class. Django does not seem
        to provide hooks for this purpose, so we simply wrap the view and
        substitute some data.
        """
        form = context['adminform'].form

        # infer correct data from the form
        fieldsets = [(None, {'fields': form.fields.keys()})]
        adminform = helpers.AdminForm(form, fieldsets, self.prepopulated_fields)
        media = mark_safe(self.media + adminform.media)

        context.update(adminform=adminform, media=media)

        return super(BaseEntityAdmin, self).render_change_form(request, context, **kwargs)


class BaseSchemaAdmin(ModelAdmin):

    list_display = ('title', 'name', 'datatype', 'help_text', 'required')
    prepopulated_fields = {'name': ('title',)}



# -*- coding: utf-8 -*-
#
#    EAV-Django is a reusable Django application which implements EAV data model
#    Copyright © 2009—2010  Andrey Mikhaylenko
#
#    This file is part of EAV-Django.
#
#    EAV-Django is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    EAV-Django is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with EAV-Django.  If not, see <http://gnu.org/licenses/>.

# TODO: .filter(size__isnull=True) --> .exclude(attrs__schema='size')

from django.db.models import Manager


RANGE_INTERSECTION_LOOKUP = 'overlaps'


class BaseEntityManager(Manager):

    # TODO: refactor filter() and exclude()   -- see django.db.models.manager and ...query

    def exclude(self, *args, **kw):
        qs = self.get_query_set().exclude(*args)
        for lookup, value in kw.items():
            lookups = self._filter_by_lookup(qs, lookup, value)
            qs = qs.exclude(**lookups)
        return qs

    def filter(self, *args, **kw):
        """
        A wrapper around standard filter() method. Allows to construct queries
        involving both normal fields and EAV attributes without thinking about
        implementation details. Usage::

            ConcreteEntity.objects.filter(rubric=1, price=2, colour='green')

        ...where `rubric` is a ForeignKey field, and `colour` is the name of an
        EAV attribute represented by Schema and Attr models.
        """

        qs = self.get_query_set().filter(*args)
        for lookup, value in kw.items():
            lookups = self._filter_by_lookup(qs, lookup, value)
            qs = qs.filter(**lookups)
        return qs

    def _filter_by_lookup(self, qs, lookup, value):

        # TODO: refactor (make recursive resolving of sublookups)

        fields   = self.model._meta.get_all_field_names()
        schemata = dict((s.name, s) for s in self.model.get_schemata_for_model())

        if '__' in lookup:
            name, sublookup = lookup.split('__', 1)
        else:
            name, sublookup = lookup, None

        if name in fields:
            try:
                related_model = getattr(self.model, name).related.model
            except AttributeError:
                pass
            else:
                # check if sublookup is another schema
                # TODO: handle nested sublookups (probably these blocks should be taken out of the Manager)
                related_schemata = dict((s.name, s) for s in related_model.get_schemata_for_model())
                if '__' in sublookup:
                    subname, subsublookup = sublookup.split('__', 1)
                else:
                    subname, subsublookup = sublookup, None
                if subname in related_schemata:
                    # EAV attribute (Attr instance linked to entity)
                    schema = related_schemata.get(subname)
                    if schema.datatype == schema.TYPE_MANY:
                        d = self._filter_by_m2m_schema(qs, subname, subsublookup, value, schema, model=related_model)
                    elif schema.datatype == schema.TYPE_RANGE:
                        d = self._filter_by_range_schema(qs, subname, subsublookup, value, schema)
                    else:
                        d = self._filter_by_simple_schema(qs, subname, subsublookup, value, schema)
                    prefixed = dict(('%s__%s' % (name, k), v) for k, v in d.items())
                    #assert 1==0, (schema, prefixed)
                    return prefixed
            # okay, treat as ordinary model field
            return {lookup: value}

        elif name in schemata:
            # EAV attribute (Attr instance linked to entity)
            schema = schemata.get(name)
            if schema.datatype == schema.TYPE_MANY:
                return self._filter_by_m2m_schema(qs, name, sublookup, value, schema)
            elif schema.datatype == schema.TYPE_RANGE:
                return self._filter_by_range_schema(qs, name, sublookup, value, schema)
            else:
                return self._filter_by_simple_schema(qs, lookup, sublookup, value, schema)
        else:
            raise NameError('Cannot filter items by attributes: unknown '
                            'attribute "%s". Available fields: %s. '
                            'Available schemata: %s.' % (name,
                            ', '.join(fields), ', '.join(schemata)))

    def _filter_by_simple_schema(self, qs, lookup, sublookup, value, schema):
        """
        Filters given entity queryset by an attribute which is linked to given
        schema and has given value in the field for schema's datatype.
        """
        value_lookup = 'attrs__value_%s' % schema.datatype
        if sublookup:
            value_lookup = '%s__%s' % (value_lookup, sublookup)
        return {
            'attrs__schema': schema,
            str(value_lookup): value
        }

    def _filter_by_range_schema(self, qs, lookup, sublookup, value, schema):
        """
        Filters given entity queryset by an attribute which is linked to given
        range schema. Lookups `between` yields items that lie not completely
        within given range but have intersection with it. For example, and item
        with x=(2,5) will match q=(3,None) or q=(0,3). See tests for details.

            qs.filter(weight_range__between=(1,3))
            qs.filter(weight_range__between=(1,None))

        """
        # This code was written with a single use case in mind. That use case
        # required searching for objects whose ranges *intersect* with given
        # one. I did not invest time in supporting other use cases (such as
        # checking whether the ranges are exactly the same or how they are
        # different). However, such lookups *can* be added later without
        # breaking existing client code. Patches are welcome.
        sublookup = sublookup or RANGE_INTERSECTION_LOOKUP
        if not sublookup == RANGE_INTERSECTION_LOOKUP:
            raise ValueError('Range schema only supports lookup "%s".' %
                             RANGE_INTERSECTION_LOOKUP)
        try:
            _, _ = value
        except ValueError:
            raise ValueError('Range schema value must be a tuple of min and '
                             'max values; one of them may be None.')
        value_lookups = zip((
            'attrs__value_range_max__gte',
            'attrs__value_range_min__lte',
        ), value)
        conditions = dict((k,v) for k,v in value_lookups if v is not None)
        conditions.update({
            'attrs__schema': schema,
        })
        return conditions

    def _filter_by_m2m_schema(self, qs, lookup, sublookup, value, schema, model=None):
        """
        Filters given entity queryset by an attribute which is linked to given
        many-to-many schema.
        """
        model = model or self.model
        schemata = dict((s.name, s) for s in model.get_schemata_for_model())   # TODO cache this dict, see above too
        try:
            schema = schemata[lookup]
        except KeyError:
            # TODO: smarter error message, i.e. how could this happen and what to do
            raise ValueError(u'Could not find schema for lookup "%s"' % lookup)
        sublookup = '__%s'%sublookup if sublookup else ''
        return {
            'attrs__schema': schema,
            'attrs__choice%s'%sublookup: value,  # TODO: can we filter by id, not name?
        }

    def create(self, **kwargs):
        """
        Creates entity instance and related Attr instances.

        Note that while entity instances may filter schemata by fields, that
        filtering does not take place here. Attribute of any schema will be saved
        successfully as long as such schema exists.

        Note that we cannot create attribute with no pre-defined schema because
        we must know attribute type in order to properly put value into the DB.
        """

        fields = self.model._meta.get_all_field_names()
        schemata = dict((s.name, s) for s in self.model.get_schemata_for_model())

        # check if all attributes are known
        possible_names = set(fields) | set(schemata.keys())
        wrong_names = set(kwargs.keys()) - possible_names
        if wrong_names:
            raise NameError('Cannot create %s: unknown attribute(s) "%s". '
                            'Available fields: (%s). Available schemata: (%s).'
                            % (self.model._meta.object_name, '", "'.join(wrong_names),
                               ', '.join(fields), ', '.join(schemata)))

        # init entity with fields
        instance = self.model(**dict((k,v) for k,v in kwargs.items() if k in fields))

        # set attributes; instance will check schemata on save
        for name, value in kwargs.items():
            setattr(instance, name, value)

        # save instance and EAV attributes
        instance.save(force_insert=True)

        return instance

'''
class BaseSchemaManager(Manager):

    def for_form(self, *args, **kw):
        return self.filter(choices=None, *args, **kw)

    def for_lookups(self, *args, **kw):
        return self.filter(datatype__not=self.model.TYPE_MANY, *args, **kw)
'''



# -*- coding: utf-8 -*-
#
#    EAV-Django is a reusable Django application which implements EAV data model
#    Copyright © 2009—2010  Andrey Mikhaylenko
#
#    This file is part of EAV-Django.
#
#    EAV-Django is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    EAV-Django is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with EAV-Django.  If not, see <http://gnu.org/licenses/>.

# python
from itertools import chain

# django
from django.db import models
from django import forms
from django.utils.datastructures import SortedDict
from django.utils.translation import ugettext as _

# 3rd-party
from view_shortcuts.decorators import cached_property

# this app
from fields import RangeField


class Facet(object):
    def __init__(self, facet_set, schema=None, field=None, lookup_prefix=''):
        self.facet_set = facet_set
        assert schema or field, 'Facet must be created with schema or field'
        assert not (schema and field), 'Facet cannot be created with both schema and field'
        self.schema = schema
        self.field = field
        self.lookup_prefix = lookup_prefix

    def __repr__(self):
        return u'<%s: %s>' % (self.__class__.__name__, unicode(self))

    def __unicode__(self):
        return unicode(self.schema.title if self.schema else self.field.verbose_name)

    extra = {}

    @property
    def field_class(self):
        raise NotImplementedError('Facet subclasses must specify form field class')

    @property
    def widget(self):
        return self.field_class.widget

    @property
    def form_field(self):
        "Returns appropriate form field."
        defaults = dict(required=False, label=unicode(self), widget=self.widget)
        defaults.update(self.extra)
        return self.field_class(**defaults)

    @property
    def attr_name(self):
        "Returns attribute name for this facet"
        return self.schema.name if self.schema else self.field.name

    @property
    def lookup_name(self):
        return '%s%s' % (self.lookup_prefix, self.attr_name)

    def get_lookups(self, value):
        "Returns dictionary of lookups for facet-specific query."
        return {self.lookup_name: value} if value else {}


class TextFacet(Facet):
    field_class = forms.ChoiceField

    @property
    def _choices(self):
        if self.schema:
            # FIXME implementation details exposed ###########
            attrs = self.schema.attrs.all()    # FIXME in shop don't need *all* attrs, just those in rubric
            field_name = 'value_%s' % self.schema.datatype
        else:
            attrs = self.facet_set.get_queryset()
            field_name = self.attr_name
        choices = set(attrs.values_list(field_name, flat=True).distinct())
        return [('', _('any'))] + [(x,x) for x in choices]

    @property
    def extra(self):
        d = {'choices': self._choices}
        if len(self._choices) < 5:
            d['widget'] = forms.RadioSelect
        return d


class ManyToManyFacet(Facet):
    field_class = forms.models.ModelMultipleChoiceField

    def _get_queryset(self):
        assert self.schema.datatype == self.schema.TYPE_MANY
        # TODO: intersection with entity and, maybe, FacetSet?
        return self.schema.get_choices()

    @property
    def extra(self):
        return {
            'queryset': self._get_queryset(),
            'widget': forms.CheckboxSelectMultiple,
        }

    def get_lookups(self, value):
        "Returns dictionary of lookups for facet-specific query."
        return {'%s__in' % self.lookup_name: value} if value else {}


class IntegerFacet(Facet):
    field_class = forms.IntegerField


class RangeFacet(Facet):
    "A simple range facet: two widgets, one attribute value (number)."
    field_class = RangeField

    def get_lookups(self, value):
        if not value:
            return {}
        # XXX what about __lt ?
        start, stop = value
        if not start:
            return {'%s__gt' % self.lookup_name: start}
        return {'%s__range' % self.lookup_name: (start or 0, stop)}


class MultiRangeFacet(Facet):
    "A complex range facet: two widgets, two attribute values (numbers)."
    field_class = RangeField

    def get_lookups(self, value):
        # we assume that this goes through the custom manager's filter()
        return {self.lookup_name: value} if value else {}

class DateFacet(Facet):
    field_class = forms.DateField


class BooleanFacet(Facet):
    field_class = forms.NullBooleanField

    # XXX this is funny but using RadioSelect for booleans is non-trivial
    #widget = RadioSelect

    def get_lookups(self, value):
        return {self.lookup_name: value} if value is not None else {}


FACET_FOR_DATATYPE_DEFAULTS = {
    'text':  TextFacet,
    'float': RangeFacet, #IntegerFacet,
    'range': MultiRangeFacet,
    'date':  DateFacet,
    'bool':  BooleanFacet,
    'many':  ManyToManyFacet,
}

FACET_FOR_FIELD_DEFAULTS = {
    models.FloatField: RangeFacet,
}


class BaseFacetSet(object):
    filterable_fields = []
    sortable_fields = []

    def __getitem__(self, k):
        return self.object_list[k]

    def __init__(self, data):
        self.data = data

    def __iter__(self):
        return iter(self.object_list)

    def __len__(self):
        if self.object_list:
            return self.object_list.count()
        return 0

    def get_queryset(self, **kwargs):
        raise NotImplementedError('BaseFacetSet subclasses must define get_queryset()')

    def get_schemata(self):
        return self.get_queryset().model.get_schemata_for_model()

    @cached_property
    def filterable_schemata(self):
        return self.get_schemata().filter(filtered=True)

    @cached_property
    def sortable_schemata(self):
        return self.get_schemata().filter(sortable=True)

    @cached_property
    def _schemata_by_name(self):
        return dict((s.name, s) for s in chain(self.filterable_schemata, self.sortable_schemata))

    def get_schema(self, name):
        return self._schemata_by_name[name]

    @cached_property
    def filterable_names(self):
        return self.filterable_fields + [str(s.name) for s in self.filterable_schemata]

    @cached_property
    def sortable_names(self):
        return self.sortable_fields + [str(s.name) for s in self.sortable_schemata]

    def _get_facets(self):
        # TODO: refactor this

        for name in self.filterable_names:
            try:
                schema, lookup_prefix = self.get_schema_and_lookup(name)
            except KeyError:    # XXX  are we sure it will raise KeyError? depends on implementation!
                field, lookup_prefix = self.get_field_and_lookup(name)
                FacetClass = FACET_FOR_FIELD_DEFAULTS.get(type(field), TextFacet)
                yield FacetClass(self, field=field, lookup_prefix=lookup_prefix)
            else:
                FacetClass = FACET_FOR_DATATYPE_DEFAULTS[schema.datatype]
                yield FacetClass(self, schema=schema, lookup_prefix=lookup_prefix)

    @cached_property
    def facets(self):
        return list(self._get_facets())

    @cached_property
    def form(self):
        if not hasattr(self, '_form'):
            fields = SortedDict([(facet.attr_name, facet.form_field) for facet in self.facets])
            FormClass = type('%sForm' % self.__class__.__name__, (forms.Form,), fields)        # XXX maybe add rubric slug?
            self._form = FormClass(self.data)
        return self._form

    def get_field_and_lookup(self, name):
        """
        Returns field instance and lookup prefix for given attribute name.
        Can be overloaded in subclasses to provide filtering across multiple models.
        """
        name = self.get_queryset().model._meta.get_field(name)
        lookup_prefix = ''
        return name, lookup_prefix

    def get_schema_and_lookup(self, name):
        schema = self.get_schema(name)
        lookup_prefix = ''
        return schema, lookup_prefix

    def get_lookups(self):
        lookups = {}
        for facet in self.facets:
            data  = self.form[facet.attr_name].data
            field = self.form.fields[facet.attr_name]
            value = field.clean(data)
            lookups.update(facet.get_lookups(value))
        return lookups

    @cached_property
    def object_list(self):
        try:
            lookups = self.get_lookups()
        except forms.ValidationError:
            return []
        lookups = dict((str(k),v) for k,v in lookups.items())

        # assume to use the EntityManager's smart filter()
        qs = self.get_queryset(**lookups).distinct()

        order_by_name = self.data.get('order_by')
        if order_by_name:
            qs = self.sort_by_attribute(qs, order_by_name)
        return qs


    def sort_by_attribute(self, qs, name):
        """
        A wrapper around standard order_by() method. Allows to sort by both normal
        fields and EAV attributes without thinking about implementation details.
        Usage::

            qs = sort_by_attributes(qs, 'price', 'colour')

        ...where `price` is a FloatField, and `colour` is the name of an EAV attribute
        represented by Schema and Attr models.
        """
        fields   = self.get_queryset().model._meta.get_all_field_names()
        schemata = self.sortable_names
        direction = '-' if self.data.get('order_desc') else ''
        if name in fields:
            return qs.order_by('%s%s' % (direction, name))
        elif name in schemata:
            schema = self.get_schema(name)
            value_lookup = 'attrs__value_%s' % schema.datatype
            order_lookup = '%s%s' % (direction, value_lookup)
            return qs.filter(attrs__schema__name=name).order_by(order_lookup)
        else:
            raise NameError('Cannot order items by attributes: unknown '
                            'attribute "%s". Available fields: %s. '
                            'Available schemata: %s.' % (name,
                            ', '.join(fields), ', '.join(schemata)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    EAV-Django is a reusable Django application which implements EAV data model
#    Copyright © 2009—2010  Andrey Mikhaylenko
#
#    This file is part of EAV-Django.
#
#    EAV-Django is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    EAV-Django is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with EAV-Django.  If not, see <http://gnu.org/licenses/>.


import os
from setuptools import setup


readme = open(os.path.join(os.path.dirname(__file__), 'README')).read()

setup(
    # overview
    name     = 'eav-django',
    description  = ('a reusable Django application which implements the '
                    'Entity-Attribute-Value data model.'),
    long_description = readme,

    # technical info
    version  = '1.1.0',
    packages = ['eav'],
    requires = ['python (>= 2.5)', 'django (>= 1.1)',
                'django_autoslug (>= 1.3.9)'],
    provides = ['eav'],

    # copyright
    author       = 'Andrey Mikhaylenko',
    author_email = 'andy@neithere.net',
    license      = 'GNU Lesser General Public License (LGPL), Version 3',

    # more info
    url          = 'http://bitbucket.org/neithere/eav-django/',
    download_url = 'http://bitbucket.org/neithere/eav-django/get/tip.zip',

    # categorization
    keywords     = 'django eav flexible data model object entity attribute value',
    classifiers  = [
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
        'Programming Language :: Python',
        'Topic :: Database',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Text Processing :: General',
    ],
)



# -*- coding: utf-8 -*-
#
#    EAV-Django is a reusable Django application which implements EAV data model
#    Copyright © 2009—2010  Andrey Mikhaylenko
#
#    This file is part of EAV-Django.
#
#    EAV-Django is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    EAV-Django is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with EAV-Django.  If not, see <http://gnu.org/licenses/>.

"""

##
## basic EAV
##

>>> colour = Schema.objects.create(title='Colour', datatype=Schema.TYPE_TEXT)
>>> colour
<Schema: Colour (text)>
>>> colour.name              #  <-- automatically generated from title
'colour'
>>> taste = Schema.objects.create(title='Taste', datatype=Schema.TYPE_TEXT)
>>> age = Schema.objects.create(title='Age', datatype=Schema.TYPE_FLOAT)
>>> can_haz = Schema.objects.create(title='I can haz it', datatype=Schema.TYPE_BOOLEAN)
>>> can_haz.name
'i_can_haz_it'
>>> e = Entity.objects.create(title='Apple', colour='green')
>>> e.title
'Apple'
>>> e.colour
'green'
>>> e.attrs.all()
[<Attr: Apple: Colour "green">]
>>> e.taste = 'sweet'
>>> e.attrs.all()
[<Attr: Apple: Colour "green">]
>>> e.colour = 'yellow'
>>> e.save()
>>> e.attrs.all()
[<Attr: Apple: Colour "yellow">, <Attr: Apple: Taste "sweet">]
>>> [x for x in e]
[<Attr: Apple: Colour "yellow">, <Attr: Apple: Taste "sweet">]
>>> Entity.objects.filter(title='Apple')
[<Entity: Apple>]
>>> Entity.objects.filter(colour='yellow')
[<Entity: Apple>]
>>> Entity.objects.filter(colour='yellow', title='Apple')
[<Entity: Apple>]

##
## range
##

>>> weight_range = Schema.objects.create(name='weight_range',
...                                      title='Weight range',
...                                      datatype=Schema.TYPE_RANGE)
>>> e = Entity.objects.all()[0]    # reload schemata cache

# try setting wrong values

>>> e.weight_range = 1
>>> e.save()
Traceback (most recent call last):
...
TypeError: Range value must be an iterable, got "1".


>>> e.weight_range = 1, 2, 3
>>> e.save()
Traceback (most recent call last):
...
ValueError: Range value must consist of two elements, got 3.

>>> e.weight_range = 1, 'wrong type'
>>> e.save()
Traceback (most recent call last):
...
TypeError: Range value must consist of two numbers, got "1" and "wrong type" instead.

>>> e.weight_range = 3, 1
>>> e.save()
Traceback (most recent call last):
...
ValueError: Range must consist of min and max values (min <= max) but got "3" and "1" instead.

# okay, now set a correct value

>>> e.weight_range = 1, 3
>>> e.save()

>>> Attr.objects.all().order_by('schema', 'choice__id')
[<Attr: Apple: Colour "yellow">, <Attr: Apple: Taste "sweet">,\
 <Attr: Apple: Weight range "(1.0, 3.0)">]

# check if queries work

>>> Entity.objects.filter(weight_range__overlaps=(1, 4))
[<Entity: Apple>]
>>> Entity.objects.filter(weight_range__overlaps=(0, None))
[<Entity: Apple>]
>>> Entity.objects.filter(weight_range__overlaps=(None, 5))
[<Entity: Apple>]
>>> Entity.objects.filter(weight_range__overlaps=(0, 0))
[]
>>> Entity.objects.filter(weight_range__overlaps=(None, 0))
[]
>>> Entity.objects.filter(weight_range__overlaps=(4, None))
[]
>>> Entity.objects.filter(weight_range__overlaps=(-5, 0))
[]
>>> Entity.objects.filter(weight_range__overlaps=(-5, 1))
[<Entity: Apple>]


##
## many-to-one
##

>>> size = Schema.objects.create(name='size', title='Size', datatype=Schema.TYPE_MANY)
>>> small  = size.choices.create(title='S')
>>> medium = size.choices.create(title='M')
>>> large  = size.choices.create(title='L')
>>> small
<Choice: S>
>>> large.schema
<Schema: Size (multiple choices)>
>>> e = Entity(title='T-shirt')
>>> e.size = small
>>> e.save()
>>> e2 = Entity.objects.get(pk=e.pk)
>>> e2.size
[<Choice: S>]
>>> e2.size = [medium, large]
>>> e2.save()
>>> e3 = Entity.objects.get(pk=e.pk)
>>> e3.size
[<Choice: M>, <Choice: L>]
>>> Attr.objects.all().order_by('schema', 'choice__id')
[<Attr: Apple: Colour "yellow">, <Attr: T-shirt: Size "M">,\
 <Attr: T-shirt: Size "L">, <Attr: Apple: Taste "sweet">,\
 <Attr: Apple: Weight range "(1.0, 3.0)">\
]
>>> e2.size = ['wrong choice']
>>> e2.save()
Traceback (most recent call last):
    ...
TypeError: Cannot assign "wrong choice": "Attr.choice" must be a BaseChoice instance.
>>> e2.size = [small, large]
>>> e2.save()
>>> e3 = Entity.objects.get(pk=e.pk)
>>> e3.size
[<Choice: S>, <Choice: L>]
>>> Attr.objects.all().order_by('schema', 'choice__id')
[<Attr: Apple: Colour "yellow">, <Attr: T-shirt: Size "S">,\
 <Attr: T-shirt: Size "L">, <Attr: Apple: Taste "sweet">,\
 <Attr: Apple: Weight range "(1.0, 3.0)">\
]

##
## combined
##

>>> Entity.objects.create(title='Orange', colour='orange', taste='sweet', size=medium)
<Entity: Orange>
>>> Entity.objects.create(title='Tangerine', colour='orange', taste='sweet', size=small)
<Entity: Tangerine>
>>> Entity.objects.create(title='Old Dog', colour='orange', taste='bitter', size=large)
<Entity: Old Dog>

>>> Entity.objects.filter(taste='sweet')
[<Entity: Apple>, <Entity: Orange>, <Entity: Tangerine>]
>>> Entity.objects.filter(colour='orange', size__in=[small, large])
[<Entity: Tangerine>, <Entity: Old Dog>]

#
# exclude() fetches objects that either have given attribute(s) with other values
# or don't have any attributes for this schema at all:
#
>>> Entity.objects.exclude(size=small)
[<Entity: Apple>, <Entity: Orange>, <Entity: Old Dog>]
>>> Entity.objects.exclude(taste='sweet')
[<Entity: T-shirt>, <Entity: Old Dog>]
>>> Entity.objects.filter(size=large) & Entity.objects.exclude(colour='orange')
[<Entity: T-shirt>]
>>> Entity.objects.filter(size=large) & Entity.objects.filter(colour='orange')
[<Entity: Old Dog>]

##
## facets
##

# make some schemata available for filtering entities by them
>>> Schema.objects.filter(name__in=['colour', 'size', 'taste']).update(filtered=True)
3
>>> fs = FacetSet({})
>>> fs.filterable_schemata
[<Schema: Colour (text)>, <Schema: Size (multiple choices)>, <Schema: Taste (text)>]
>>> fs.filterable_names
['price', 'colour', 'size', 'taste']
>>> fs.facets
[<TextFacet: Item price>, <TextFacet: Colour>, <ManyToManyFacet: Size>, <TextFacet: Taste>]
>>> [x for x in fs]
[<Entity: Apple>, <Entity: T-shirt>, <Entity: Orange>, <Entity: Tangerine>, <Entity: Old Dog>]
>>> [x for x in FacetSet({'colour': 'yellow'})]
[<Entity: Apple>]
>>> [x for x in FacetSet({'colour': 'orange'})]
[<Entity: Orange>, <Entity: Tangerine>, <Entity: Old Dog>]
>>> [x for x in FacetSet({'colour': 'orange', 'taste': 'sweet'})]
[<Entity: Orange>, <Entity: Tangerine>]
>>> [x for x in FacetSet({'size': [large.pk]})]
[<Entity: T-shirt>, <Entity: Old Dog>]
"""

# TODO: if schema changes type, drop all attribs?

# django
from django.contrib.contenttypes import generic
from django.db import models

# this app
from facets import BaseFacetSet
from models import BaseAttribute, BaseChoice, BaseEntity, BaseSchema


class Schema(BaseSchema):
    pass


class Choice(BaseChoice):
    schema = models.ForeignKey(Schema, related_name='choices')


class Attr(BaseAttribute):
    #entity = models.ForeignKey(Entity, related_name='attrs')
    schema = models.ForeignKey(Schema, related_name='attrs')
    choice = models.ForeignKey(Choice, related_name='attrs', null=True)


class Entity(BaseEntity):
    title = models.CharField(max_length=100)
    price = models.IntegerField(blank=True, null=True, verbose_name='Item price')
    attrs = generic.GenericRelation(Attr, object_id_field='entity_id',
                                    content_type_field='entity_type')

    @classmethod
    def get_schemata_for_model(cls):
        return Schema.objects.all()

    def __unicode__(self):
        return self.title


class FacetSet(BaseFacetSet):
    filterable_fields = ['price']
    sortable_fields = ['price']

    def get_queryset(self, **kwargs):
        return Entity.objects.filter(**kwargs)     # can be pre-filtered using custom FacetSet.__init__
